/*
 *  lettertree.c
 *  yalex
 *
 *  Created by Benot Sagot on Fri Feb 27 2004.
 *  Copyright (c) 2004 INRIA. All rights reserved.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lettertree.h"

#define MAX_LINE_LENGTH 255
#define NODE_CHILD_BLOCK 4
#define LEAF_DATA_BLOCK 4

#ifdef DEBUG
#define debug(format, args...) fprintf (stderr, format, args)
#else
#define debug(format, args...) 
#endif

Node * new_tree(void) {
	return calloc(1, sizeof(Node));
}

void _delete_tree(Node * tree) {
	int i;

	/* recurse depth first in children */
	for (i = 0; i < tree->nd; i++)
		_delete_tree(tree->down + i);

	debug("node %c\n", tree->c);

	/* free leaf */
	free(tree->leaf);

	/* free children, if they exists */
	if (tree->nd)
		free(tree->down);
}

void delete_tree(Node * tree) {
	/* delete tree recursively */
	_delete_tree(tree);

	/* delete root */
	free(tree);
}

int add_data(Node * tree, char * word, void * data) {
	Node *currnode;
	char *c;
	int i;
	Leaf *leaf;

	debug("Adding word \"%s\"\n", word);

	currnode = tree;
	for (c = word; *c != 0; c++) {
		if (currnode->nd == 0) {
			debug(">%c", *c);
			currnode->nd = 1;
			currnode->down = malloc(NODE_CHILD_BLOCK * sizeof(Node));
			currnode = currnode->down;
			currnode->c = *c;
			currnode->nd = 0;
			currnode->leaf = NULL;
		} else {
			for (i = 0; i < currnode->nd && *c != currnode->down[i].c; i++) ;
			if (i == currnode->nd) {
				debug("+%c", *c);
				if (currnode->nd % NODE_CHILD_BLOCK == 0)
					currnode->down = realloc(currnode->down, (NODE_CHILD_BLOCK + currnode->nd) * sizeof(Node));
				currnode->nd++;
				currnode = currnode->down + i;
				currnode->c = *c;
				currnode->nd = 0;
				currnode->leaf = NULL;
			} else {
				currnode = currnode->down + i;
				debug(" %c", *c);
			}
		}
	}

	if (currnode->leaf == NULL) {
		currnode->leaf = malloc(sizeof(Leaf));
		leaf = currnode->leaf;
		leaf->data = malloc((LEAF_DATA_BLOCK + 1) * sizeof(void*));
		leaf->count = 0;
	} else {
		leaf = currnode->leaf;
		if (leaf->count % LEAF_DATA_BLOCK == 0)
			leaf->data = realloc(leaf->data, (LEAF_DATA_BLOCK + currnode->leaf->count + 1) * sizeof(void*));
	}
	leaf->data[leaf->count] = data;
	leaf->data[leaf->count + 1] = NULL;
	leaf->count++;

	debug("\n", NULL);

	return 0;
}

int _parse_tree(Node * tree, char * s) {
	int i;
	char se[MAX_LINE_LENGTH];

	if (tree->c != '\0')
		sprintf(se, "%s%c", s, tree->c);
	else
		sprintf(se, "%c", tree->c);

	if (tree->leaf != NULL)
		printf("%s\t%s\n", se, tree->leaf->data);

	if (tree->nd != 0) {
		for (i = 0; i < tree->nd; i++)
			_parse_tree(tree->down + i, se);
	}

	return 0;
}

int parse_tree(Node * tree) {
	char se[MAX_LINE_LENGTH];
	_parse_tree(tree, se);
}

Node * _find_node(Node * tree, char * word) {
	Node *currnode;
	char *c;
	int i;

	debug("Looking for \"%s\"\n", word);

	currnode = tree;
	for (c = word; *c != 0; c++) {
		if (currnode->nd == 0) {
			/* no more child */
			debug(">%c\n",*c);
			return NULL;
		} else {
			for (i = 0; i < currnode->nd && *c != currnode->down[i].c; i++) ;
			if (i == currnode->nd) {
				/* no correct child */
				debug("+%c\n", *c);
				return NULL;
			} else {
				/* traverse child */
				debug(" %c", *c);
				currnode = currnode->down + i;
			}
		}
	}

	debug("\n", NULL);

	return currnode;
}

int has_word(Node * tree, char * word) {
	Node *node = _find_node(tree, word);

	/* return leaf existence */
	return node && node->leaf != NULL;
}

void** get_data(Node * tree, char * word) {
	Node *node = _find_node(tree, word);

	/* return leaf content */
	return node && node->leaf != NULL ?
		node->leaf->data :
		NULL;
}
